from unittest.mock import AsyncMock

import pytest
from fastapi import HTTPException, status

import provisioning_plugin.dependencies
from id_broker_common.kelvin import kelvin_session
from provisioning_plugin.models import SchoolClass
from provisioning_plugin.routes.v1 import school_classes, schools

from ..utils import (
    CLASS_SIZE,
    DEFAULT_SCHOOL_AUTHORITY,
    KelvinSessionMock,
    get_preconfigured_session_mock,
    kelvin_school,
    kelvin_school_class,
    kelvin_student,
)


def test_school_by_name(func_client, broker_school):
    _client = func_client(
        custom_overrides={
            provisioning_plugin.dependencies.get_school_by_id: kelvin_school,
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    response = _client.get(url=f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/schools/{broker_school.name}")
    assert response.status_code == 200
    assert response.json() == broker_school


@pytest.mark.parametrize("ou_exists", (True, False))
def test_school_get_by_name(func_client, broker_school, ou_exists: bool):
    async def _get_school_by_id():  # -> KelvinSchool
        print("*** _get_school_by_id() ***")
        if ou_exists:
            return kelvin_school()
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    _client = func_client(
        custom_overrides={
            provisioning_plugin.dependencies.get_school_by_id: _get_school_by_id,
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    ou_name = broker_school.name if ou_exists else "fake_ou123"  # TODO
    response = _client.get(url=f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/schools/{ou_name}")
    assert response.status_code == 200 if ou_exists else 404
    if ou_exists:
        assert response.json() == broker_school


@pytest.mark.parametrize("ou_exists", (True, False))
def test_school_exists_by_name(func_client, broker_school, ou_exists: bool):
    async def _find_school() -> str:
        if ou_exists:
            return broker_school.name
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    _client = func_client(
        custom_overrides={
            provisioning_plugin.dependencies.find_school_name_by_id_in_primary_ldap: _find_school,
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    ou_name = broker_school.name if ou_exists else "fake_ou123"  # TODO
    response = _client.head(url=f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/schools/{ou_name}")
    assert response.status_code == 200 if ou_exists else 404


def test_post_school(func_client, broker_school, monkeypatch, async_iterable, mocker):
    _client = func_client(
        custom_overrides={
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    monkeypatch.setattr(
        schools,
        "get_pseudonyms_for_providers",
        async_iterable,
    )
    kelvinSessionMock = KelvinSessionMock()
    kelvinSessionMock.configure_mock(
        **{
            "post.return_value": kelvin_school(),
            "__aenter__.return_value": kelvinSessionMock,
        }
    )
    mocker.patch(
        "provisioning_plugin.routes.v1.schools.IDBrokerKelvinSession",
        lambda *args, **kwargs: kelvinSessionMock,
    )
    response = _client.post(f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/schools", json=broker_school)
    assert response.status_code == 201
    assert response.json() == broker_school


@pytest.mark.parametrize("method", ["GET", "HEAD"])
def test_school_class_by_id(func_client, monkeypatch, method, broker_school_class):
    exp = SchoolClass.with_prefix(SchoolClass(**broker_school_class.dict()), DEFAULT_SCHOOL_AUTHORITY)
    monkeypatch.setattr(SchoolClass, "from_kelvin_school_class", AsyncMock(return_value=exp))
    _client = func_client(
        custom_overrides={
            provisioning_plugin.dependencies.get_school_class_by_id: kelvin_school_class,
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    query = _client.get if method == "GET" else _client.head
    response = query(url=f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/classes/{broker_school_class.id}")
    assert response.status_code == 200
    if method == "GET":
        res = response.json()
        res["members"] = sorted(res["members"])
        assert res == broker_school_class
    else:
        assert response.status_code == 200, response.__dict__
        assert response.content == b""


@pytest.mark.parametrize(
    "method,status_code",
    [
        ("POST", 201),
        ("PUT", 200),
    ],
)
def test_post_school_class(
    func_client, monkeypatch, method, status_code, broker_school_class, async_iterable
):
    monkeypatch.setattr(
        school_classes,
        "member_ids_to_usernames",
        AsyncMock(return_value=[kelvin_student(i) for i in range(CLASS_SIZE)]),
    )
    _client = func_client(
        custom_overrides={
            kelvin_session: get_preconfigured_session_mock(
                {
                    "post.return_value": kelvin_school_class(),
                }
            ),
            provisioning_plugin.dependencies.get_school_class_by_id: kelvin_school_class,
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    monkeypatch.setattr(
        school_classes,
        "get_pseudonyms_for_providers",
        async_iterable,
    )
    url = f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/classes"
    query = None
    if method == "POST":
        query = _client.post
    elif method == "PUT":
        query = _client.put
        url += f"/{broker_school_class.id}"
    response = query(
        url,
        json=broker_school_class,
    )
    print(response.__dict__)
    assert response.status_code == status_code
    res = response.json()
    res["members"] = sorted(res["members"])
    assert res == broker_school_class


@pytest.mark.parametrize(
    "method",
    [
        "POST",
        "PUT",
    ],
)
def test_put_post_school_class_missing_member(
    func_client, monkeypatch, method, broker_school_class, async_iterable
):
    class ldap_access:
        async def search(self, filter_s, **kwargs):
            return []

    monkeypatch.setattr(
        provisioning_plugin.dependencies,
        "ldap_access",
        ldap_access,
    )

    _client = func_client(
        custom_overrides={
            kelvin_session: get_preconfigured_session_mock(
                {
                    "post.return_value": kelvin_school_class(),
                }
            ),
            provisioning_plugin.dependencies.get_school_class_by_id: kelvin_school_class,
            provisioning_plugin.dependencies.check_for_authority_admin: lambda: True,
        }
    )
    monkeypatch.setattr(
        school_classes,
        "get_pseudonyms_for_providers",
        async_iterable,
    )
    url = f"/v1/{DEFAULT_SCHOOL_AUTHORITY}/classes"
    query = None
    change = "created"
    if method == "POST":
        query = _client.post
    elif method == "PUT":
        query = _client.put
        url += f"/{broker_school_class.id}"
        change = "modified"
    response = query(
        url,
        json=broker_school_class,
    )
    assert response.status_code == 422, response.json()
    assert response.json()["detail"] in [
        f"The school class '{broker_school_class['name']}' in school "
        f"'{DEFAULT_SCHOOL_AUTHORITY}-{broker_school_class['school']}' in school authority "
        f"'{DEFAULT_SCHOOL_AUTHORITY}' could not be {change}: "
        f"\"Object '{member}' in school authority '{DEFAULT_SCHOOL_AUTHORITY}' not found.\""
        for member in broker_school_class["members"]
    ]
